#ifndef BattlePlan_H
#define BattlePlan_H

#include "GameObj.H"

using namespace std;

struct Objective {
	int minResources;
	int no;
	int done;
	int cmd;
	sint4 executer;
};

/*
	This class holds a battleplan to be used through a game.
*/
class BattlePlan {

	typedef GameObj* GameObjPtr;
	
	public:

		/*
			Constructor
		*/
		BattlePlan();

		/*
			Adds a new objective to the plan. The objective is placed last in the queue.

			minResources -  Minimal number of resources when this objective shall be executed.
			no - Number of times to execute this objective.
			cmd - Objective command, see constants below.
			execute - The structure responsible for executing the objective, for example factory
		*/
		void addObjective(int minResources, int no, int cmd, sint4 executer);

		/*
			Same as addObjective but the objective is added first in list instead of last.
		*/
		void insertObjectiveFirst(int minResources, int no, int cmd, sint4 executer);

		/*
			Returns a matching objective based on the current number of resources
			available. If no objective is found, and Objective with command NONE_FOUND
			is returned.
		*/
		Objective getMatchingObjective(int resources, sint4 executer);

		/*
			Tells the plan that the current objective has been executed.
		*/
		void objectiveExecuted(sint4 executer);

		/*
			Returns the number of objectives in the plan.
		*/
		int noObjectives();

		//Objectives command constants.

		//No matching objective found. Do nothing.
		static const sint4 NONE_FOUND;
		//Build barracks
		static const sint4 BUILD_BARRACKS;
		//Build a factory
		static const sint4 BUILD_FACTORY;
		//Build [no] number of tanks
		static const sint4 BUILD_TANK;
		//Train [no] number of workers
		static const sint4 TRAIN_WORKER;
		//Train [no] number of marines
		static const sint4 TRAIN_MARINE;
		//Set a worker to explore
		static const sint4 EXPLORE;

		int engage_at_attackpower;
		
	private:
		
		vector<Objective> objectives;

		int noObjs;
};

#endif
