#include "CommanderInChief.H"
#include "WorkerAgent.H"
#include "Blackboard.H"
#include "Options.H"

BattlePlan* CommanderInChief::generateBattlePlan(const Game& game) {

	Blackboard::map_w = game.get_map().get_width();
	Blackboard::map_h = game.get_map().get_height();

	BattlePlan* plan = new BattlePlan();
	noWorkers = 0;
	noExplorers = 0;
	noTanks = 0;
	noMarines = 0;
	noBases = 0;
	noBarracks = 0;
	noFactories = 0;

	bool game1;
	Options::get("-game1", game1);
	if (game1) {
		//Game 1 - Resource gathering - no plan
		Blackboard::game_type = 1;
		plan->engage_at_attackpower = 1000; //Not used
		return plan;
	}

	//Default game - Game 3
	Blackboard::game_type = 3;
	plan->addObjective(0, 2, BattlePlan::EXPLORE, Blackboard::TYPE_MAIN);
	plan->addObjective(400, 1, BattlePlan::BUILD_BARRACKS, Blackboard::TYPE_MAIN);
	plan->addObjective(200, 5, BattlePlan::TRAIN_WORKER, Blackboard::TYPE_BASE);
	plan->addObjective(410, 1, BattlePlan::BUILD_FACTORY, Blackboard::TYPE_MAIN);
	plan->addObjective(200, 5, BattlePlan::TRAIN_WORKER, Blackboard::TYPE_BASE);
	plan->addObjective(200, 5, BattlePlan::BUILD_TANK, Blackboard::TYPE_FACTORY);
	plan->addObjective(200, 50, BattlePlan::BUILD_TANK, Blackboard::TYPE_FACTORY);

	plan->engage_at_attackpower = 25;

	return plan;
}

void CommanderInChief::updateBattlePlan(PFUpdaterPtr pfupdater, BattlePlan* plan) {
	int cnoWorkers = 0;
	int cnoExplorers = 0;
	int cnoTanks = 0;
	int cnoMarines = 0;
	int cnoBases = 0;
	int cnoBarracks = 0;
	int cnoFactories = 0;

	vector<ObjFuturePos>* units = pfupdater->getPF()->getUnitsList();
	vector<ObjFuturePos>::iterator it = units->begin();
	while (it != units->end()) {
		if (it->isWorker()) {
			cnoWorkers++;
	
			if(it->currentState == WorkerAgent::STATE_EXPLORE) {
				cnoExplorers++;
			}
		}
		if (it->isTank()) {
			cnoTanks++;
		}
		if (it->isMarine()) {
			cnoMarines++;
		}
		it++;
	}

	vector<BasePos>* bases = pfupdater->getPF()->getBasesList();
	vector<BasePos>::iterator bit = bases->begin();
	while (bit != bases->end()) {
		if (bit->isControlCenter()) {
			cnoBases++;
		}
		if (bit->isBarracks()) {
			cnoBarracks++;
		}
		if (bit->isFactory()) {
			cnoFactories++;
		}
		bit++;
	}

	//*** Check for changes ***

	//Structures built
	if (cnoBarracks > noBarracks) {
		Blackboard::BARRACKS_BUILT = true;
	}
	if (cnoFactories > noFactories) {
		Blackboard::FACTORY_BUILT = true;
	}
	if (cnoBases > noBases) {
		Blackboard::CONTROLCENTER_BUILT = true;
	}

	//Structures destroyed
	if (cnoBarracks < noBarracks) {
		cout << "Barracks destroyed. Build new!" << endl;
		plan->insertObjectiveFirst(410, 1, BattlePlan::BUILD_BARRACKS, Blackboard::TYPE_MAIN);
	}
	if (cnoFactories < noFactories) {
		cout << "Factory destroyed. Build new!" << endl;
		plan->insertObjectiveFirst(410, 1, BattlePlan::BUILD_FACTORY, Blackboard::TYPE_MAIN);
	}

	//Worker issues
	if (cnoExplorers < noExplorers) {
		cout << "Explorer Down! Assign new" << endl;
		plan->insertObjectiveFirst(0, 1, BattlePlan::EXPLORE, Blackboard::TYPE_MAIN);
	}
	if (cnoWorkers < noWorkers) {
		int no = noWorkers - cnoWorkers;
		cout << no << " workers down. Create new" << endl;
		plan->insertObjectiveFirst(60, no, BattlePlan::TRAIN_WORKER, Blackboard::TYPE_BASE);
	}

	//Assign
	noWorkers = cnoWorkers;
	noExplorers = cnoExplorers;
	noTanks = cnoTanks;
	noMarines = cnoMarines;
	noBases = cnoBases;
	noBarracks = cnoBarracks;
	noFactories = cnoFactories;

	/*cout << "-------- STATUS --------" << endl;
	cout << "Workers: " << noWorkers << endl;
	cout << "Explorers: " << noExplorers << endl;
	cout << "Marines: " << noMarines << endl;
	cout << "Tanks: " << noTanks << endl;
	cout << "Control centers: " << noBases << endl;
	cout << "Barracks: " << noBarracks << endl;
	cout << "Factories: " << noFactories << endl;
	cout << "------------------------" << endl;*/
}
