#ifndef CommanderInChief_H
#define CommanderInChief_H

#include "GameObj.H"
#include "BattlePlan.H"
#include "PFUpdater.H"

using namespace std;

/*
	This class represents the commander-in-chief agent. It is responsible for generating a battle plan
	to be used for the current game, and update the plan during the game if necessary.
*/
class CommanderInChief {

	typedef GameObj* GameObjPtr;
	typedef PFUpdater* PFUpdaterPtr;
	
	public:

		/*
			Generates the default battle plan.
		*/
		BattlePlan* generateBattlePlan(const Game& game);

		/*
			Updates the battle plan (if needed). Should be called about every 10:th frame or so.
		*/
		void updateBattlePlan(PFUpdaterPtr pfupdater, BattlePlan* plan);

	private:

		//Counters for game objects.
		//Used to see if, for example, an own worker is destroyed.
		int noWorkers;
		int noExplorers;
		int noTanks;
		int noMarines;
		int noBases;
		int noBarracks;
		int noFactories;

};

#endif
