#include "CommanderInField.H"
#include "WorkerAgent.H"
#include "Blackboard.H"

void CommanderInField::setBattlePlan(BattlePlan* mPlan) {
	plan = mPlan;
}

BattlePlan* CommanderInField::getBattlePlan() {
	return plan;
}

FireCoordinator* CommanderInField::getCoordinator() {
	return &coordinator;
}

void CommanderInField::compute_actions(GameStateModulePtr state, PFUpdaterPtr pfupdater, Random* rand) {
	coordinator.coordinate_fire(&*state, &*pfupdater);

	//Check how many resources we have
	Game& game = state->get_game();
	PlayerInfo& pi = game.get_cplayer_info();
	ScriptObj *so = pi.global_obj("player");
	int resources = (int)so->get_int("minerals");

	int cAttackPower = 0;
	
	//Check for MAIN objectives
	Objective cObjective = plan->getMatchingObjective(resources, Blackboard::TYPE_MAIN);

	//Check pre conditions
	bool barracks_built = false;
	if (cObjective.cmd == BattlePlan::BUILD_FACTORY) {
		barracks_built = barracksBuilt(pfupdater);
	}

 	vector<ObjFuturePos>* units = pfupdater->getPF()->getUnitsList();
	vector<ObjFuturePos>::iterator it = units->begin();
	while (it != units->end()) {
		if (it->isWorker()) {
			if (cObjective.cmd == BattlePlan::BUILD_BARRACKS && it->isAvailable()) {
				//Assign a worker to build barracks
				//Just assigns first worker to build, can be optimised more
				it->currentState = WorkerAgent::STATE_BUILD_BARRACKS;
				plan->objectiveExecuted(Blackboard::TYPE_MAIN);
				cObjective.cmd = BattlePlan::NONE_FOUND; //Clear
				cout << "[" << it->objId << "] Assigned to build barracks" << endl;
			}
			if (cObjective.cmd == BattlePlan::BUILD_FACTORY && barracks_built && it->isAvailable()) {
				//Assign a worker to build factory
				//Just assigns first worker to build, can be optimised more
				it->currentState = WorkerAgent::STATE_BUILD_FACTORY;
				plan->objectiveExecuted(Blackboard::TYPE_MAIN);
				cObjective.cmd = BattlePlan::NONE_FOUND; //Clear
				cout << "[" << it->objId << "] Assigned to build factory" << endl;
			}
			if (cObjective.cmd == BattlePlan::EXPLORE && it->isAvailable()) {
				//Assign a worker to explore
				it->currentState = WorkerAgent::STATE_EXPLORE;
				plan->objectiveExecuted(Blackboard::TYPE_MAIN);
				cObjective.cmd = BattlePlan::NONE_FOUND; //Clear
				cout << "[" << it->objId << "] Assigned to explore" << endl;
			}
			
			WorkerAgent::compute_actions(it->obj, &*state, &*pfupdater, &*rand, it->currentState);
		}
		else {
			//Not worker = combat unit
			UnitAgent::compute_actions(it->obj, &*state, &*pfupdater, &coordinator, &*rand);

			if (it->isTank()) {
				cAttackPower += 5;
			}
			if (it->isMarine()) {
				cAttackPower += 1;
			}
		}
		it++;
	}

	//Check if engage the enemy
	if (cAttackPower >= plan->engage_at_attackpower && Blackboard::state == Blackboard::DEFEND) {
		cout << "Start attack the enemy" << endl;
		Blackboard::state = Blackboard::ATTACK;
	}

	//Check for CONTROLCENTER objectives
	if (controlCenterBuilt(pfupdater)) {
		Objective cObjective = plan->getMatchingObjective(resources, Blackboard::TYPE_BASE);
	
		if (cObjective.cmd == BattlePlan::TRAIN_WORKER) {
			vector<BasePos>* bases = pfupdater->getPF()->getBasesList();
			vector<BasePos>::iterator bit = bases->begin();
			while (bit != bases->end()) {
				if (bit->isControlCenter()) {
					if (trainWorker(bit->obj)) {
						//Train worker
						//cout << "[" << bit->objId << "] Train Worker (" << cObjective.done << "/" << cObjective.no << ")" << endl;
						cout << "[" << bit->objId << "] Train Worker" << endl;
						plan->objectiveExecuted(Blackboard::TYPE_BASE);
						break;
					}
				}
				bit++;
			}
		}
	}

	//Check for BARRACKS objectives
	if (barracksBuilt(pfupdater)) {
		Objective cObjective = plan->getMatchingObjective(resources, Blackboard::TYPE_BARRACKS);
	
		if (cObjective.cmd == BattlePlan::TRAIN_MARINE) {
			vector<BasePos>* bases = pfupdater->getPF()->getBasesList();
			vector<BasePos>::iterator bit = bases->begin();
			while (bit != bases->end()) {
				if (bit->isBarracks()) {
					if (trainMarine(bit->obj)) {
						//Train worker
						//cout << "[" << bit->objId << "] Train Marine (" << cObjective.done << "/" << cObjective.no << ")" << endl;
						cout << "[" << bit->objId << "] Train Marine" << endl;
						plan->objectiveExecuted(Blackboard::TYPE_BARRACKS);
						break;
					}
				}
				bit++;
			}
		}
	}

	//Check for FACTORY objectives
	if (factoryBuilt(pfupdater)) {
		Objective cObjective = plan->getMatchingObjective(resources, Blackboard::TYPE_FACTORY);
	
		if (cObjective.cmd == BattlePlan::BUILD_TANK) {
			vector<BasePos>* bases = pfupdater->getPF()->getBasesList();
			vector<BasePos>::iterator bit = bases->begin();
			while (bit != bases->end()) {
				if (bit->isFactory()) {
					if (buildTank(bit->obj)) {
						//Train worker
						//cout << "[" << bit->objId << "] Build Tank (" << cObjective.done << "/" << cObjective.no << ")" << endl;
						cout << "[" << bit->objId << "] Build Tank" << endl;
						plan->objectiveExecuted(Blackboard::TYPE_FACTORY);
						break;
					}
				}
				bit++;
			}
		}
	}
}

bool CommanderInField::barracksBuilt(PFUpdaterPtr pfupdater) {
	vector<BasePos>* bases = pfupdater->getPF()->getBasesList();
	vector<BasePos>::iterator bit = bases->begin();
	while (bit != bases->end()) {
		if (bit->isBarracks()) {
			return true;
		}
		bit++;
	}
	return false;
}

bool CommanderInField::controlCenterBuilt(PFUpdaterPtr pfupdater) {
	vector<BasePos>* bases = pfupdater->getPF()->getBasesList();
	vector<BasePos>::iterator bit = bases->begin();
	while (bit != bases->end()) {
		if (bit->isControlCenter()) {
			return true;
		}
		bit++;
	}
	return false;
}

bool CommanderInField::factoryBuilt(PFUpdaterPtr pfupdater) {
	vector<BasePos>* bases = pfupdater->getPF()->getBasesList();
	vector<BasePos>::iterator bit = bases->begin();
	while (bit != bases->end()) {
		if (bit->isFactory()) {
			return true;
		}
		bit++;
	}
	return false;
}

bool CommanderInField::buildTank(GameObjPtr factory) {
	if(factory->get_int("amount_built") != factory->get_int("build_time")) {
		//Building not ready
		return false;
	}

	Vector<sint4> args;
	if (factory->get_int("task_done") != 0) {
		return false;
	}
	factory->set_action("build_tank", args);
	return true;
}

bool CommanderInField::trainMarine(GameObjPtr barracks) {
	if(barracks->get_int("amount_built") != barracks->get_int("build_time")){
		//Building not ready
		return false;
	}

	Vector<sint4> args;
	if (barracks->get_int("task_done") != 0) {
		return false;
	}
	barracks->set_action("train_marine", args);
	return true;
}

bool CommanderInField::trainWorker(GameObjPtr controlCenter) {
	if(controlCenter->get_int("amount_built") != controlCenter->get_int("build_time")){
		//Building not ready
		return false;
	}

	Vector<sint4> args;
	if (controlCenter->get_int("task_done") != 0) {
		return false;
	}
	controlCenter->set_action("train_worker", args);
	return true;
}
