#ifndef CommanderInField_H
#define CommanderInField_H

#include "GameObj.H"
#include "BattlePlan.H"
#include "PFUpdater.H"
#include "GameStateModule.H"
#include "FireCoordinator.H"
#include "Random.H"
#include "UnitAgent.H"
#include "WorkerAgent.H"
#include "PF.H"

using namespace std;

/*
	This class represents the commander-in-field agent. It is responsible for executing the battleplan
	generated by the commander-in-chief agent.
*/
class CommanderInField {

	typedef GameObj* GameObjPtr;
	typedef PFUpdater* PFUpdaterPtr;
	typedef GameStateModule* GameStateModulePtr;
	
	public:

		/*
			Builds a tank. Returns true if it is successful.
		*/
		static bool buildTank(GameObjPtr factory);
		
		/*
			Trains a marine. Returns true if it is successful.
		*/
		static bool trainMarine(GameObjPtr barracks);

		/*
			Trains a worker. Returns true if it is successful.
		*/
		static bool trainWorker(GameObjPtr controlCenter);

		/*
			Sets the battle plan to execute.
		*/
		void setBattlePlan(BattlePlan* mPlan);
		
		/*
			Returns the battle plan.
		*/
		BattlePlan* getBattlePlan();

		/*
			Compute actions for the commander-in-field each frame. It selects appropriate actions
			for units and workers, builds structures etc all according to the battleplan.
		*/
		void compute_actions(GameStateModulePtr state, PFUpdaterPtr pfupdater, Random* rand);

		/*
			Returns a reference to the fire coordinator agent.
		*/
		FireCoordinator* getCoordinator();

	private:
		
		BattlePlan* plan;

		/*
			Returns true if a barracks has been built. Barracks is pre-condition to building a
			factory.
		*/
		bool barracksBuilt(PFUpdaterPtr pfupdater);

		/*
			Returns true if controlCenter has been built.
		*/
		bool controlCenterBuilt(PFUpdaterPtr pfupdater);

		/*
			Returns true if a factory has been built.
		*/
		bool factoryBuilt(PFUpdaterPtr pfupdater);

		FireCoordinator coordinator;
		Random rand; // local RNG
};

#endif
