#include "PF.H"
#include "TotalPF.H"
#include "EventTimer.H"
#include "TBConfig.H"
#include "Blackboard.H"
#include "WorkerAgent.H"
#include "UnitAgent.H"

enum { TILE_UNKNOWN=0, TILE_WATER=1, TILE_GROUND=2, TILE_CLIFF=3 };

PF::PF() {
	for (int cX = 0; cX < 64; cX++) {
		for (int cY = 0; cY < 64; cY++) {
			expl_map[cX][cY].visited();
		}
	}
}

void PF::updateExplMap(const Game& game) {
	//Note: Dont visit edge points
	for (int cX = 3; cX < 61; cX++) {
		for (int cY = 3; cY < 61; cY++) {
			expl_map[cX][cY].visitCnt++;
		}
	}

	//Units
	vector<ObjFuturePos>::iterator it = units.begin();
	while (it != units.end()) {
		setTileVisited(it->x, it->y, 1);
		it++;
	}

	//Terrain
	const Map<GameTile> &gameMap = game.get_map();
	FORS (y, gameMap.get_height()) {
		FORS (x, gameMap.get_width()) {
			const GameTile &tile = gameMap(x,y);

			if (tile.get_split() == Tile::NO_SPLIT) {
				if (tile.get_typeW() == TILE_CLIFF) {
					//Impassable terrain.... dont explore
					setTileVisited(x * 16, y * 16, 3); //2
				}
			}
			else {
				//Splitted tiles always has impassable terrain... dont explore them
				setTileVisited(x * 16, y * 16, 3);
			}
		}
	}

	//Bases
	vector<BasePos>::iterator bit = bases.begin();
	while (bit != bases.end()) {
		
		int startX = bit->x1 / 16;
		int endX = bit->x2 / 16;
		int startY = bit->y1 / 16;
		int endY = bit->y2 / 16;

		for (int cX = startX; cX <= endX; cX++) {
			for (int cY = startY; cY <= endY; cY++) {
				setTileVisited(cX * 16, cY * 16, 0);
			}
		}

		bit++;
	}

	//Enemy Bases
	vector<BasePos>::iterator eit = enemyBases.begin();
	while (eit != enemyBases.end()) {
		
		int startX = eit->x1 / 16;
		int endX = eit->x2 / 16;
		int startY = eit->y1 / 16;
		int endY = eit->y2 / 16;

		for (int cX = startX; cX <= endX; cX++) {
			for (int cY = startY; cY <= endY; cY++) {
				setTileVisited(cX * 16, cY * 16, 0);
			}
		}

		eit++;
	}

	//Minerals
	vector<MineralPos>::iterator mit = minerals.begin();
	while (mit != minerals.end()) {
		setTileVisited(mit->x, mit->y, 3);
		mit++;
	}
}

Position PF::getNextToVisit(int unitX, int unitY) {
	int bestX = -1;
	int bestY = -1;
	real8 highestImportance = 0;

	Random rand;
	rand.srand((uint4)clock_t());
	
	int vCnt = 0;
	int tCnt = 0;

	for (int cX = 0; cX < 64; cX++) {
		for (int cY = 0; cY < 64; cY++) {
			real8 visitValue = (real8)expl_map[cX][cY].visitCnt * (real8)2.4;
			if (visitValue > 0) {
				int mX = cX * 16 + 8;
				int mY = cY * 16 + 8;
				real8 dist = TotalPF::distance(mX, mY, unitX, unitY);
				real8 importance = visitValue + (((real8)1500 * 0.1 - dist) / (real8)1500); // * 0.1
				
				importance += (real8)(rand.ranged_uint4(1, 8) / 10); //1-10 / 10
				
				if (importance > highestImportance) {
					//cout << "dist=" << dist << "visitValue=" << visitValue << " imp=" << importance << " highestImp=" << highestImportance << endl;
					highestImportance = importance;
					bestX = mX;
					bestY = mY;
				}
				vCnt++;
			}
			tCnt++;
		}
	}

	Position pos;
	pos.x = bestX;
	pos.y = bestY;

	//Set next position to visited as well
	//setTileVisited(bestX, bestY, 2);

	return pos;
}

void PF::setTileVisited(int x, int y, int range) {
	int tX = x / 16;
	int tY = y / 16;

	for (int cX = tX - range; cX <= tX + range; cX++) {
		for (int cY = tY - range; cY <= tY + range; cY++) {
			if (cX >= 0 && cX < 64 && cY >= 0 && cY < 64) {
				expl_map[cX][cY].visited();
			}
		}
	}
}

Position PF::getDestination(uint4 objId) {
	Position pos;
	pos.x = 0;
	pos.y = 0;

	vector<ObjFuturePos>::iterator it = units.begin();
	while (it != units.end()) {
		if ((*it).isThisObj(objId)) {
			pos.x = it->destX;
			pos.y = it->destY;
			break;
		}
		it++;
	}

	return pos;
}

void PF::clearDestination(uint4 objId) {
	vector<ObjFuturePos>::iterator it = units.begin();
	while (it != units.end()) {
		if ((*it).isThisObj(objId)) {
			it->destX = 0;
			it->destY = 0;
			return;
		}
		it++;
	}
}

void PF::setDestination(uint4 objId, Position pos) {
	vector<ObjFuturePos>::iterator it = units.begin();
	while (it != units.end()) {
		if ((*it).isThisObj(objId)) {
			it->destX = pos.x;
			it->destY = pos.y;
			return;
		}
		it++;
	}
}

sint4 PF::getCurrentState(uint4 objId) {
	vector<ObjFuturePos>::iterator it = units.begin();
	while (it != units.end()) {
		if ((*it).isThisObj(objId)) {
			return it->currentState;
		}
		it++;
	}

	return 0;
}

void PF::updateCurrentState(uint4 objId, sint4 newState) {
	vector<ObjFuturePos>::iterator it = units.begin();
	while (it != units.end()) {
		if ((*it).isThisObj(objId)) {
			it->currentState = newState;
			return;
		}
		it++;
	}
}

int PF::getOU() {
	return units.size();
}

int PF::getEU() {
	return enemies.size();
}

int PF::getOB() {
	return bases.size();
}

int PF::getEB() {
	return enemyBases.size();
}

vector<ObjFuturePos>* PF::getUnitsList() {
	return &units;
}

vector<BasePos>* PF::getBasesList() {
	return &bases;
}

vector<ObjPos>* PF::getEnemyUnitsList() {
	return &enemies;
}

vector<BasePos>* PF::getEnemyBasesList() {
	return &enemyBases;
}

vector<MineralPos>* PF::getMineralsList() {
	return &minerals;
}

void PF::setAllDestroyed() {
	vector<ObjPos>::iterator it = enemies.begin();
	while (it != enemies.end()) {
		(*it).destroyed = true;
		it++;
	}

	vector<ObjFuturePos>::iterator uit = units.begin();
	while (uit != units.end()) {
		(*uit).destroyed = true;
		uit++;
	}

	vector<MineralPos>::iterator mit = minerals.begin();
	while (mit != minerals.end()) {
		(*mit).destroyed = true;
		mit++;
	}
}

void PF::cleanUp() {
	for (int i = 0; i < (int)enemies.size(); i++) {
		if (enemies[i].destroyed) {
			vector<ObjPos>::iterator it = enemies.begin();
			it += i;
			enemies.erase(it);
			i--;
		}
	}

	for (int i = 0; i < (int)units.size(); i++) {
		if (units[i].destroyed) {
			vector<ObjFuturePos>::iterator it = units.begin();
			it += i;
			units.erase(it);
			i--;
		}
	}

	for (int i = 0; i < (int)minerals.size(); i++) {
		if (minerals[i].destroyed) {
			vector<MineralPos>::iterator it = minerals.begin();
			it += i;
			minerals.erase(it);
			i--;
		}
	}
}

void PF::updateMineralPosition(uint4 objId, int x, int y) {
	//Check if struct for this mineral exists and, if found, update
	bool found = false;
	vector<MineralPos>::iterator it = minerals.begin();
	while (it != minerals.end()) {
		if ((*it).isThisObj(objId)) {
			(*it).destroyed = false;
			found = true;
		}
		it++;
	}

	//Mineral struct not found. Add new
	if (!found) {
		MineralPos pos;
		pos.objId = objId;
		pos.x = x;
		pos.y = y;
		minerals.push_back(pos);
	}
}

real8 PF::getMineralCharge(int pX, int pY, uint4 objId, int unitType, int marg, bool attractive) { //objId for future use
	real8 p = 0;

	int baseRadius = 3; //Worker radius
	if (unitType == Blackboard::TYPE_MARINE) {
		baseRadius = 4;
	}
	if (unitType == Blackboard::TYPE_TANK) {
		baseRadius = 10; //7
	}
	baseRadius += marg;

	vector<MineralPos>::iterator it = minerals.begin();
	while (it != minerals.end()) {
		if (!it->destroyed) {
			if (attractive) {
				real8 tmp = calcMineralCharge(pX, pY, it->x, it->y, attractive, baseRadius);
				if (tmp > p) {
					p = tmp;
					closestMineral = *it;
				}
			}
			else {
				real8 tmp = calcMineralCharge(pX, pY, it->x, it->y, attractive, baseRadius);
				p += tmp;
			}
		}
		it++;
	}
	
	return p;
}

real8 PF::calcMineralCharge(int x, int y, int objX, int objY, bool attractive, int baseRadius) {
	real8 p = 0;
	real8 dist = TotalPF::distance(x, y, objX, objY);

	if (attractive) {
		p = (real8)240 - (dist) * 0.32;
		if (p < 0) p = 0;
	}
	else {
		int radius = baseRadius + 6; //mineral radius 6
		p = calcUnitCharge(dist, radius);
	}

	//TODO Introduce weights later
	//p = p * TBConfig::params.Minerals_w;
	
	return p;
}

void PF::updateSheepPosition(uint4 game_tick, uint4 objId, int x, int y) {
	//Check if struct for this unit exists and, if found, update
	bool found = false;
	vector<ObjPos>::iterator it = sheep.begin();
	while (it != sheep.end()) {
		if ((*it).isThisObj(objId)) {
			(*it).update(game_tick, x, y);
			found = true;
		}
		it++;
	}

	//Unit struct not found. Add new
	if (!found) {
		ObjPos pos;
		pos.objId = objId;
		pos.x = x;
		pos.y = y;
		pos.gametick = game_tick;
		sheep.push_back(pos);
	}
}

real8 PF::getSheepCharge(int pX, int pY, uint4 game_tick, int baseRadius) {
	real8 p = 0;

	vector<ObjPos>::iterator it = sheep.begin();
	while (it != sheep.end()) {
		if (TotalPF::MHdistance(pX, pY, (*it).x, (*it).y) <= 26) {
			p += calcSheepCharge(pX, pY, (*it).x, (*it).y, baseRadius);
		}

		it++;
	}
	
	return p;
}

real8 PF::calcSheepCharge(int x, int y, int objX, int objY, int baseRadius) {
	real8 dist = TotalPF::distance(x, y, objX, objY);

	real8 p = calcUnitCharge(dist, baseRadius + 4); //4 = sheep radius
	p = p * TBConfig::params.Sheep_w;
	
	return p;
}

void PF::updateEnemyUnitPosition(uint4 game_tick, uint4 objId, sint4 objType, int x, int y) {
	//Check if struct for this unit exists and, if found, update
	bool found = false;
	vector<ObjPos>::iterator it = enemies.begin();
	while (it != enemies.end()) {
		if (it->isThisObj(objId)) {
			it->update(game_tick, x, y);
			found = true;
			return;
		}
		it++;
	}

	//Unit struct not found. Add new
	if (!found) {
		ObjPos pos;
		pos.objId = objId;
		pos.type = objType;
		pos.x = x;
		pos.y = y;
		pos.gametick = game_tick;
		enemies.push_back(pos);
	}
}

real8 PF::getEnemyUnitCharge(int pX, int pY, bool offensive, int unitType, uint4 game_tick) {
	real8 p = 0;

	vector<ObjPos>::iterator it = enemies.begin();
	while (it != enemies.end()) {
		if (!(*it).destroyed && unitType == Blackboard::TYPE_MARINE) {
			if (offensive) {
				real8 mp = calcOffensiveMarineCharge(pX, pY, (*it).x, (*it).y);
				if (TBConfig::hasAddition1()) {
					if (p < mp) {
						p = mp;
					}
				}
				else {
					p += mp;
				}
			}
			if (!offensive) {
				p += calcDefensiveMarineCharge(pX, pY, (*it).x, (*it).y);
			}
		}

		if (!(*it).destroyed && unitType == Blackboard::TYPE_TANK) {
			if (offensive) {
				real8 mp = calcOffensiveTankCharge(pX, pY, (*it).x, (*it).y);
				if (TBConfig::hasAddition1()) {
					if (p < mp) {
						p = mp;
					}
				}
				else {
					p += mp;
				}
			}
			if (!offensive) {
				p += calcDefensiveTankCharge(pX, pY, (*it).x, (*it).y);
			}
		}

		if (!(*it).destroyed && unitType == Blackboard::TYPE_WORKER) {
			if (it->type == Blackboard::TYPE_TANK) {
				p += calcDefensiveTankCharge(pX, pY, (*it).x, (*it).y);
			}
			if (it->type == Blackboard::TYPE_MARINE) {
				p += calcDefensiveMarineCharge(pX, pY, (*it).x, (*it).y);
			}
			//No defensive field around enemy workers
		}
		it++;
	}
	
	return p;
}

real8 PF::calcOffensiveMarineCharge(int x, int y, int objX, int objY) {
	real8 p = 0;
	real8 dist = TotalPF::distance(x, y, objX, objY);
	
	real8 fadeVal = 0.12;
	if (Blackboard::state == Blackboard::DEFEND) {
		fadeVal *= 5;
	}

	//TODO Tune
	if (dist <= 24) {
		p = (real8)325 / (real8)24 * dist;
	}
	else if (dist <= 26) {
		p = 325;
	}
	else {
		p = (real8)240 - (dist - 26) * fadeVal;
		if (p < 0) p = 0;
	}
	
	return p;
}

real8 PF::calcDefensiveMarineCharge(int x, int y, int objX, int objY) {
	if (TotalPF::MHdistance(x, y, objX, objY) > 144) return 0;
	real8 p = 0;
	real8 dist = TotalPF::distance(x, y, objX, objY);
	if (dist <= 72) { //72  64=max shoot dist
		p = (-60 + (dist * 0.8)) * 10;
	}
	
	return p;
}

real8 PF::calcOffensiveTankCharge(int x, int y, int objX, int objY) {
	real8 p = 0;
	real8 dist = TotalPF::distance(x, y, objX, objY);

	real8 fadeVal = 0.24;
	if (Blackboard::state == Blackboard::DEFEND) {
		fadeVal *= 3;
	}

	if (dist <= 110) {
		p = (real8)240 / (real8)110 * dist;
	}
	else if (dist > 110 && dist <= 112) {
		p = 240;
	}
	else {
		p = 240 - (dist - 112) * fadeVal;
		if (p < 0) p = 0;
	}
	p = p * TBConfig::params.OffTank_w;
	
	return p;
}

real8 PF::calcDefensiveTankCharge(int x, int y, int objX, int objY) {
	real8 dist = TotalPF::distance(x, y, objX, objY);
	real8 p = -800 + dist * 6.4;
	if (p > 0) {
		p = 0;
	}
	p = p * TBConfig::params.DefTank_w;
	
	return p;
}

void PF::updateUnitPosition(uint4 game_tick, GameObj* obj, uint4 objId, sint4 objType, int x, int y) {
	//Check if struct for this unit exists and, if found, update
	bool found = false;
	vector<ObjFuturePos>::iterator it = units.begin();
	while (it != units.end()) {
		if ((*it).isThisObj(objId)) {
			(*it).update(game_tick, x, y);
			found = true;

			//Update trail
			(*it).trail.update(x, y, game_tick);
			
			return;
		}
		it++;
	}

	//Unit struct not found. Add new
	if (!found) {
		ObjFuturePos pos;
		pos.objId = objId;
		pos.obj = obj;
		pos.type = objType;
		if (objType == Blackboard::TYPE_WORKER) {
			pos.currentState = WorkerAgent::STATE_MOVE_TO_MINE;
			pos.miscState = 0;
		}
		pos.x = x;
		pos.y = y;
		pos.destX = 0;
		pos.destY = 0;
		pos.gametick = game_tick;
		units.push_back(pos);

		//Init trail
		Trail trail;
		trail.init(x, y, game_tick);
		pos.trail = trail;
	}
}

void PF::setNextUnitPosition(uint4 objId, int nextX, int nextY) {
	vector<ObjFuturePos>::iterator it = units.begin();
	while (it != units.end()) {
		if ((*it).isThisObj(objId)) {
			(*it).nextX = nextX;
			(*it).nextY = nextY;
			return;
		}
		it++;
	}
}

real8 PF::getUnitCharge(int pX, int pY, uint4 objId, int unitType, uint4 game_tick) {
	real8 p = 0;

	int baseRadius = 3; //Worker radius
	if (unitType == Blackboard::TYPE_MARINE) {
		baseRadius = 4;
	}
	if (unitType == Blackboard::TYPE_TANK) {
		baseRadius = 7;
	}

	vector<ObjFuturePos>::iterator it = units.begin();
	while (it != units.end()) {
		if (!(*it).destroyed && (*it).isMarine() && (*it).objId != objId) {
			p += calcMarineUnitCharge(pX, pY, (*it).x, (*it).y, baseRadius);
			if ((*it).hasNextPos()) {
				p += calcMarineUnitCharge(pX, pY, (*it).nextX, (*it).nextY, baseRadius);
			}
		}
		if (!(*it).destroyed && (*it).isTank() && (*it).objId != objId) {
			p += calcTankUnitCharge(pX, pY, (*it).x, (*it).y, baseRadius);
			if ((*it).hasNextPos()) {
				p += calcTankUnitCharge(pX, pY, (*it).nextX, (*it).nextY, baseRadius);
			}
		}
		if (!(*it).destroyed && (*it).isWorker() && (*it).objId != objId) {
			p += calcWorkerUnitCharge(pX, pY, (*it).x, (*it).y, baseRadius);
			if ((*it).hasNextPos()) {
				p += calcWorkerUnitCharge(pX, pY, (*it).nextX, (*it).nextY, baseRadius);
			}
		}

		if (TBConfig::hasAddition4()) {
			//Trail charge
			if (it->isThisObj(objId)) {
				p += it->trail.getP(pX, pY, game_tick);
			}
		}

		it++;
	}
	
	return p;
}

real8 PF::calcUnitCharge(real8 dist, int radius) {
	real8 p = 0;
	if (dist <= radius) {
		p = - 20;
	}
	else {
		real8 mp = (radius + 3) - 2 * (dist - radius);
		if (mp < 0) mp = 0;

		p = p - mp;
	}
	
	return p;
}

real8 PF::calcMarineUnitCharge(int x, int y, int objX, int objY, int baseRadius) {
	if (TotalPF::MHdistance(x, y, objX, objY) > 20) return 0;

	int radius = baseRadius + 4; //4 = marine radius
	real8 dist = TotalPF::distance(x, y, objX, objY);
	
	return calcUnitCharge(dist, radius);
}

real8 PF::calcWorkerUnitCharge(int x, int y, int objX, int objY, int baseRadius) {
	if (TotalPF::MHdistance(x, y, objX, objY) > 20) return 0;

	int radius = baseRadius + 3; //3 = worker radius
	real8 dist = TotalPF::distance(x, y, objX, objY);
	
	return calcUnitCharge(dist, radius);
}

real8 PF::calcTankUnitCharge(int x, int y, int objX, int objY, int baseRadius) {
	if (TotalPF::MHdistance(x, y, objX, objY) > 22) return 0;

	int radius = baseRadius + 7; //7 = worker radius
	real8 dist = TotalPF::distance(x, y, objX, objY);
	
	return calcUnitCharge(dist, radius);
}

void PF::clearBaseLists() {
	bases.clear();
	enemyBases.clear();
}

void PF::addBase(uint4 objId, GameObj* obj, int x, int y, int x1, int y1, int x2, int y2, bool own, int type) {
	//First, check if bases exists
	if (own) {
		vector<BasePos>::iterator it = bases.begin();
		while (it != bases.end()) {
			if (it->x == x && it->y == y) {
				return;
			}
			it++;
		}
	}
	else {
		vector<BasePos>::iterator it = enemyBases.begin();
		while (it != enemyBases.end()) {
			if (it->x == x && it->y == y) {
				return;
			}
			it++;
		}
	}

	BasePos pos;
	pos.objId = objId;
	pos.obj = obj;
	pos.x = x;
	pos.y = y;
	pos.x1 = x1;
	pos.y1 = y1;
	pos.x2 = x2;
	pos.y2 = y2;
	pos.type = type;
	if (own) {
		bases.push_back(pos);
	}
	else {
		enemyBases.push_back(pos);
	}
}

void PF::removeBase(int x, int y) {
	vector<BasePos>::iterator it = bases.begin();
	while (it != bases.end()) {
		if (it->x == x && it->y == y) {
			bases.erase(it);
			return;
		}
		it++;
	}

	vector<BasePos>::iterator eit = enemyBases.begin();
	while (eit != enemyBases.end()) {
		if (eit->x == x && eit->y == y) {
			enemyBases.erase(eit);
			return;
		}
		eit++;
	}
}

void PF::dumpBases() {
	cout << "---- Bases ----" << endl;
	vector<BasePos>::iterator it = bases.begin();
	while (it != bases.end()) {
		cout << "O id=" << it->objId << " x=" << it->x << " y=" << it->y << endl;
		it++;
	}

	vector<BasePos>::iterator eit = enemyBases.begin();
	while (eit != enemyBases.end()) {
		cout << "E id=" << eit->objId << " x=" << eit->x << " y=" << eit->y << endl;
		eit++;
	}
	cout << "---------------" << endl;
}

real8 PF::getBaseCharge(int pX, int pY, int radius, bool attractive) {
	real8 p = 0;

	vector<BasePos>::iterator it = bases.begin();
	while (it != bases.end()) {
		if (attractive) {
			if (it->isControlCenter()) {
				p += calcBaseCharge(pX, pY, (*it).x, (*it).y, (*it).isPositionInside(pX, pY, radius), true);
			}
			else {
				p += calcBaseCharge(pX, pY, (*it).x, (*it).y, (*it).isPositionInside(pX, pY, radius), false);
			}
		}
		else {
			p += calcBaseCharge(pX, pY, (*it).x, (*it).y, (*it).isPositionInside(pX, pY, radius), false);
		}
		it++;
	}

	return p;
}

real8 PF::getDefendBaseCharge(int pX, int pY, int unitType) {
	real8 p = 0;

	//TODO NOT SUM and change type

	vector<BasePos>::iterator it = bases.begin();
	while (it != bases.end()) {
		real8 mp = calcEnemyBaseCharge(pX, pY, (*it).x, (*it).y, (*it).isPositionInside(pX, pY), unitType); //TODO: Dynamic radius of circle?
		if (p == 0) {
			p = mp;
		}
		else if (mp > p) {
			p = mp;
		}
		it++;
	}

	return p * 0.01;
}

real8 PF::calcBaseCharge(int x, int y, int objX, int objY, bool inside, bool attractive) {
	real8 p = 0;

	if (!attractive) {
		if (inside) {
			p = TotalPF::CHARGE_IMPASSABLE_TILES;
		}
		else {
			real8 dist = TotalPF::distance(x, y, objX, objY);
	
			//Base width = 62
			//p = (real8)(6 * dist - 258) * 0.01;
			p = (real8)(0.5 * dist - 174) * 0.1;
			if (p > 0) p = 0;
		}
		p = p * TBConfig::params.Base_w;
	}
	else {
		real8 dist = TotalPF::distance(x, y, objX, objY);
		p = (real8)240 - (dist) * 0.32;
		if (p < 0) p = 0;
	}

	return p;
}

real8 PF::getEnemyBaseCharge(int pX, int pY, int unitType, bool attack) {
	real8 p = 0;
	
	vector<BasePos>::iterator it = enemyBases.begin();
	while (it != enemyBases.end()) {
		if (!attack) {
			p += calcBaseCharge(pX, pY, (*it).x, (*it).y, (*it).isPositionInside(pX, pY), false);
		}
		else {
			real8 mp = calcEnemyBaseCharge(pX, pY, (*it).x, (*it).y, (*it).isPositionInside(pX, pY), unitType);
			if (TBConfig::hasAddition1()) {
				if (mp > p) {
					p = mp;
				}
			}
			else {
				p += mp;
			}
		}
		it++;
	}

	return p;
}

real8 PF::calcEnemyBaseCharge(int x, int y, int objX, int objY, bool inside, int unitType) {
	real8 p = 0;
	if (inside) {
		p = 0;
	}
	else {
		real8 dist = TotalPF::distance(x, y, objX, objY) - (real8)31; //Base width = 64
		
		real8 fadeVal = 0.12;
		if (Blackboard::state == Blackboard::DEFEND) {
			fadeVal *= 5;
		}

		if (unitType == Blackboard::TYPE_TANK) {
			if (dist <= 110) {
				p = (real8)360 / (real8)110 * dist;
			}
			else if (dist <= 112) {
				p = 360;
			}
			else {
				p = (real8)240 - (dist - 112) * fadeVal;
				if (p < 0) p = 0;
			}
		}
		if (unitType == Blackboard::TYPE_MARINE) {
			//TODO Tune
			if (dist <= 24) {
				p = (real8)360 / (real8)24 * dist;
			}
			else if (dist <= 26) {
				p = 360;
			}
			else {
				p = (real8)240 - (dist - 26) * fadeVal;
				if (p < 0) p = 0;
			}
		}
		
		p = p * TBConfig::params.OffBase_w;
	}
	return p;
}

real8 PF::getBuilderBaseCharge(int pX, int pY, uint4 objId, uint4 game_tick, bool useTrail) {
	real8 p = 0;

	vector<BasePos>::iterator it = bases.begin();
	while (it != bases.end()) {
		real8 p1 = calcBuilderBaseCharge(pX, pY, it->x, it->y, it->isPositionInside(pX, pY, 2));
		p += p1;
		if (p1 <= TotalPF::CHARGE_IMPASSABLE_TILES) {
			return p1;
		}
		it++;
	}

	vector<BasePos>::iterator eit = enemyBases.begin();
	while (eit != enemyBases.end()) {
		real8 p1 = calcBuilderEnemyBaseCharge(pX, pY, eit->x, eit->y, eit->isPositionInside(pX, pY, 2));
		p += p1;
		if (p1 <= TotalPF::CHARGE_IMPASSABLE_TILES) {
			return p1;
		}
		eit++;
	}

	vector<MineralPos>::iterator mit = minerals.begin();
	while (mit != minerals.end()) {
		real8 p1 = calcBuilderMineralCharge(pX, pY, mit->x, mit->y);
		p += p1;
		mit++;
	}

	if (TBConfig::hasAddition4() && useTrail) {
		vector<ObjFuturePos>::iterator it = units.begin();
		while (it != units.end()) {
			if (!it->destroyed && it->isThisObj(objId)) {
				p += it->trail.getP(pX, pY, game_tick);
				break;
			}
			it++;
		}
	}

	return p;
}

real8 PF::calcBuilderBaseCharge(int x, int y, int objX, int objY, bool inside) {
	real8 p = 0;

	if (inside) {
		p = TotalPF::CHARGE_IMPASSABLE_TILES;
		return p;
	}
	else {
		real8 dist = TotalPF::distance(x, y, objX, objY);

		//62 (base width) * 2 + 16
		if (dist < 115) { //repelling
			return TotalPF::CHARGE_IMPASSABLE_TILES;
		}
		else {
			p = (real8)(115 - (dist - 115));
			if (p < 0) p = 0;
		}
	}
	
	return p;
}

real8 PF::calcBuilderEnemyBaseCharge(int x, int y, int objX, int objY, bool inside) {
	real8 p = 0;

	if (inside) {
		p = TotalPF::CHARGE_IMPASSABLE_TILES;
		return p;
	}
	else {
		real8 dist = TotalPF::distance(x, y, objX, objY);

		//62 (base width) * 2 + 16
		if (dist < 150) { //repelling
			return TotalPF::CHARGE_IMPASSABLE_TILES;
		}
		else {
			p = (real8)(50 - (dist - 150));
			if (p < 0) p = 0;
		}
	}
	
	return p;
}

real8 PF::calcBuilderMineralCharge(int x, int y, int objX, int objY) {
	real8 p = 0;

	real8 dist = TotalPF::distance(x, y, objX, objY);

	//62 (base width) * 2 + 16
	if (dist <= 90) { //repelling
		return TotalPF::CHARGE_IMPASSABLE_TILES;
	}
	else {
		p = (real8)(50 - dist * 0.2) * 0.1;
		if (p < 0) p = 0;
	}
	
	return p * 0.1;
}

real8 PF::getBuilderEdgeCharge(int pX, int pY) {
	if (pX >= 934) {
		return TotalPF::CHARGE_IMPASSABLE_TILES;
	}
	
	if (pY >= 934) {
		return TotalPF::CHARGE_IMPASSABLE_TILES;
	}

	if (pX < 90) {
		return TotalPF::CHARGE_IMPASSABLE_TILES;
	}
	if (pY < 90) {
		return TotalPF::CHARGE_IMPASSABLE_TILES;
	}

	return 0;
}


