#include "TBConfig.H"

using namespace std;

bool config_init = false;

Params TBConfig::params;
string TBConfig::path = "NAN";
string TBConfig::paramsString = "1#0#1#1#2829#1429#1429#1429#1429#1429#1429";

/*
	Sets path to the config file to use.
*/
void TBConfig::setConfigPath(string confpath) {
	TBConfig::path = confpath;
	cout << "Using config " << TBConfig::path << endl;
}

/*
	Sets the app params as a string instead of read from file.
*/
void TBConfig::setConfigString(string mParamsString) {
	TBConfig::paramsString = mParamsString;
}

/*
	Returns true if config is initiated, false otherwise.
*/
bool TBConfig::isInitiated() {
	return config_init;
}

/*
	Inits the config by reading a config file and set the
	parameters in the params struct.
*/
void TBConfig::init() {
	//Set default value to params struct
	params.OffTank_w = 0.2829;
	params.DefTank_w = 0.1429;
	params.Tank_w = 0.1429;
	params.Base_w = 0.1429;
	params.Sheep_w = 0.1429;
	params.OffTank_w = 0.1429;
	params.Terrain_w = 0.1429;

	params.TotalPF_PathMinThreshold = -50;
	params.TotalPF_PathMaxThreshold = -1;
	params.Attack_DmgProbability = 0.75;
	params.Addition1 = true;
	params.Addition2 = false;
	params.Addition3 = true;
	params.Addition4 = true;

	if (TBConfig::path != "NAN") {
		cout << "Reading TankBattle config file" << endl;

		if (!TBConfig::path.empty()) {
			//Read config file
			string line;
			ifstream myfile(TBConfig::path.c_str());
			if (myfile.is_open()) {
				while (!myfile.eof()) {
					getline(myfile,line);
					parseLine(line);
				}
				myfile.close();
			}
			else {
				cout << "WARNING! Config file not found! Using default values" << endl; 
			}
		}
		else {
			cout << "WARNING! Config file not set. Using default values" << endl;
		}
	}
	if (TBConfig::paramsString != "NAN") {
		cout << "Setting params from string " << TBConfig::paramsString << endl;

		vector<string> tokens;
		int no = split(TBConfig::paramsString, "#", tokens);
		if (no == 11) {
			params.Addition1 = stringToBool(tokens[0]);
			params.Addition2 = stringToBool(tokens[1]);
			params.Addition3 = stringToBool(tokens[2]);
			params.Addition4 = stringToBool(tokens[3]);
			params.OffTank_w = stringToReal(tokens[4]) / (real8)10000;
			params.DefTank_w = stringToReal(tokens[5]) / (real8)10000;
			params.Tank_w = stringToReal(tokens[6]) / (real8)10000;
			params.Base_w = stringToReal(tokens[7]) / (real8)10000;
			params.Sheep_w = stringToReal(tokens[8]) / (real8)10000;
			params.OffBase_w = stringToReal(tokens[9]) / (real8)10000;
			params.Terrain_w = stringToReal(tokens[10]) / (real8)10000;
		}
		else {
			cout << "WARNING! Invalid amount of tokens in params string. Found " << no << ", expected 11. Using default values" << endl;
		}

		cout << "Sheep_w=" << params.Sheep_w << " base_w=" << params.Base_w << " tank_w=" << params.Tank_w << " defTank_w=" << params.DefTank_w << " offTank_w=" << params.OffTank_w << " offBase=" << params.OffBase_w << " terrain=" << params.Terrain_w << endl;
	}

	config_init = true;

	cout << "Done reading config file" << endl;
}

/*
	Parses a line from the config file and, if it is a parameter line,
	sets the value in the params struct.
*/
void TBConfig::parseLine(string line) {
	if (line.find("#") != 0 && line.length() > 2) {
		vector<string> tokens;
		int no = split(line, "=", tokens);

		if (no == 2) {
			//Set param values to struct
			if (contains(line, "Addition1")) {
				params.Addition1 = stringToBool(tokens[1]);
			}
			else if (contains(line, "Addition2")) {
				params.Addition2 = stringToBool(tokens[1]);
			}
			else if (contains(line, "Addition3")) {
				params.Addition3 = stringToBool(tokens[1]);
			}
			else if (contains(line, "Addition4")) {
				params.Addition4 = stringToBool(tokens[1]);
			}
			else {
				cout << "Unknown config parameter: " << tokens[0] << endl;
			}
		}
		else {
			cout << "Invalid format for parameter: " << line << endl;
		}
	}
}

bool TBConfig::hasAddition1() {
	return params.Addition1;
}

bool TBConfig::hasAddition2() {
	return params.Addition2;
}

bool TBConfig::hasAddition3() {
	return params.Addition3;
}

bool TBConfig::hasAddition4() {
	return params.Addition4;
}

bool TBConfig::contains(string line, string token) {
	if (line.find(token) == 0) {
		return true;
	}
	return false;
}

/*
	Converts a string to a real value.
	Returns 0 if string is not a valid real.
*/
real8 TBConfig::stringToReal(string str) {
	real8 value(0);
	std::istringstream iss(str);

	if (iss >> value) {
		return value;
	}
	else {
		return 0;
	}
}

/*
	Converts a string to an int value.
	Returns 0 if string is not a valid int.
*/
int TBConfig::stringToInt(string str) {
	int value(0);
	std::istringstream iss(str);

	if (iss >> value) {
		return value;
	}
	else {
		return 0;
	}
}

/*
	Converts a string to a bool. Returns false
	if not a valid string.
*/
bool TBConfig::stringToBool(string str) {
	if (str.find("true") != string::npos) return true;
	if (str.find("TRUE") != string::npos) return true;
	if (str.find("1") != string::npos) return true;
	return false;
}

/*
	Splits a string using a specified delimiter. The tokens are stored in a
	results vector.
*/
int TBConfig::split(const std::string &input, const std::string &delimiter, std::vector<string> &results) {
    // find the start character of the next token.

	int offset = 0;
	int found = 0;
	bool end = false;

	while (!end) {
		size_t i = input.find_first_not_of(delimiter, offset);
		if (i == string::npos) {
			//No more tokens. End.
			offset = input.length();
			end = true;
			break;
		}

		// find the end of the token.
		size_t j = input.find_first_of(delimiter, i);
		if (j == string::npos) {
			string token = input.substr(i);
			offset = input.length();
			
			if (token.length() > 0) {
				results.push_back(token);
				found++;
			}
		}
		else {
			// to intercept the token and save current position
			string token = input.substr(i, j - i);
			offset = j;
			
			if (token.length() > 0) {
				results.push_back(token);
				found++;
			}
		}
	}

	return found;
}
