#include "TotalPF.H"
#include "Game.H"
#include "Blackboard.H"
#include "TBConfig.H"
#include "WorkerAgent.H"

using namespace std;

const int TotalPF::CHARGE_IMPASSABLE_TILES = -1000;

//Pre-calculated square-root constants to speed up distance calculations.
const real8 TotalPF::preCalc[101] = {
1.00, 1.00, 1.00, 1.001, 1.001, 1.001, 1.002, 1.003, 1.004, 1.004, 1.005, 1.006, 1.008, 1.009, 1.01, 1.012, 1.013, 1.015, 1.017, 1.019, 1.021, 1.023, 1.025, 1.027, 1.029, 1.032, 1.034, 1.037, 1.04, 1.042, 1.045, 1.048, 1.051, 1.054, 1.058, 1.061, 1.064, 1.068, 1.071, 1.075, 1.079, 1.082, 1.086, 1.09, 1.094, 1.098, 1.103, 1.107, 1.111, 1.116, 1.12, 1.125, 1.129, 1.134, 1.139, 1.143, 1.148, 1.153, 1.158, 1.163, 1.169, 1.174, 1.179, 1.184, 1.19, 1.195, 1.201, 1.206, 1.212, 1.218, 1.223, 1.229, 1.235, 1.241, 1.247, 1.253, 1.259, 1.265, 1.271, 1.277, 1.284, 1.29, 1.296, 1.303, 1.309, 1.316, 1.322, 1.329, 1.335, 1.342, 1.349, 1.355, 1.362, 1.369, 1.376, 1.383, 1.39, 1.397, 1.404, 1.411, 1.414
};

enum { TILE_UNKNOWN=0, TILE_WATER=1, TILE_GROUND=2, TILE_CLIFF=3 };

TotalPF::TotalPF(GameStateModulePtr mState) {
	
}

TotalPF::~TotalPF() {
	
}

real8 TotalPF::getUnitP(sint4 objX, sint4 objY, bool offensive, uint4 unitId, int unitType, const Game& game, PF* pf) {
	const Map<GameTile> &gameMap = game.get_map();
	int mX = objX / game.get_tile_points();
	int mY = objY / game.get_tile_points();
	const GameTile &tile = gameMap(mX, mY);
	if (tile.get_typeW() == TILE_UNKNOWN) {
		return 0;
	}
	
	real8 p = getTerrainCharge(objX, objY, unitType, false, game);
	if (p <= TotalPF::CHARGE_IMPASSABLE_TILES) {
		return p;
	}

	int baseRadius = 3; //Worker radius
	if (unitType == Blackboard::TYPE_MARINE) {
		baseRadius = 4;
	}
	if (unitType == Blackboard::TYPE_TANK) {
		baseRadius = 10; //7
	}

	p += pf->getSheepCharge(objX, objY, game.tick(), baseRadius);
	p += pf->getEnemyUnitCharge(objX, objY, offensive, unitType, game.tick());
	p += pf->getUnitCharge(objX, objY, unitId, unitType, game.tick());
	p += pf->getMineralCharge(objX, objY, unitId, unitType, 4, false);

	if (Blackboard::state == Blackboard::ATTACK) {
		p += pf->getBaseCharge(objX, objY, baseRadius, false) * 3;
		p += pf->getDefendBaseCharge(objX, objY, unitType);

		//Test
		//p += getUnexplCharge(objX, objY, unitId, pf) * 0.01;
	}
	else {
		p += pf->getDefendBaseCharge(objX, objY, unitType);
	}
	p += pf->getEnemyBaseCharge(objX, objY, unitType, true);
	
	return p;
}

real8 TotalPF::getWorkerP(sint4 objX, sint4 objY, sint4 currentState, uint4 unitId, int unitType, const Game& game, PF* pf, bool useTrail) {
	const Map<GameTile> &gameMap = game.get_map();
	int mX = objX / game.get_tile_points();
	int mY = objY / game.get_tile_points();
	
	const GameTile &tile = gameMap(mX, mY);
	
	real8 p = getTerrainCharge(objX, objY, unitType, false, game);
	if (p <= TotalPF::CHARGE_IMPASSABLE_TILES) {
		return p;
	}

	p += pf->getSheepCharge(objX, objY, game.tick(), 3);
	p += pf->getUnitCharge(objX, objY, unitId, unitType, game.tick());
	p += pf->getEnemyUnitCharge(objX, objY, false, unitType, game.tick());
	p += pf->getEnemyBaseCharge(objX, objY, unitType, false);
	
	//Explorer workers	
	if (currentState == WorkerAgent::STATE_EXPLORE) {
		p += pf->getBaseCharge(objX, objY, 3, false);
		p += getUnexplCharge(objX, objY, unitId, pf);
		p += pf->getMineralCharge(objX, objY, unitId, unitType, 4, false);
	}
	else if (currentState == WorkerAgent::STATE_MOVE_TO_MINE) {
		p += pf->getBaseCharge(objX, objY, 2, false);
		p += pf->getMineralCharge(objX, objY, unitId, unitType, 0, true);
	}
	else if (currentState == WorkerAgent::STATE_MOVE_TO_BASE) {
		p += pf->getMineralCharge(objX, objY, unitId, unitType, 0, false);
		p += pf->getBaseCharge(objX, objY, 2, true);
	}
	else if (currentState == WorkerAgent::STATE_BUILD_BARRACKS || currentState == WorkerAgent::STATE_BUILD_FACTORY) {
		p += pf->getBaseCharge(objX, objY, 3, false);
		p += getBuildSpotCharge(objX, objY, unitId, pf);
		p += pf->getMineralCharge(objX, objY, unitId, unitType, 3, false);
	}
	
	return p;
}

real8 TotalPF::getBuildSpotCharge(sint4 objX, sint4 objY, uint4 unitId, PF* pf) {
	real8 p = 0;
	
	Position pos = pf->getDestination(unitId);
	if (pos.x > 0 && pos.y > 0) {
		real8 dist = distance(objX, objY, pos.x, pos.y);
		p = 150 - dist * 0.1;
		if (p < 0) {
			p = 0;
		}
	}

	return p;
}

real8 TotalPF::getUnexplCharge(sint4 objX, sint4 objY, uint4 unitId, PF* pf) {
	Position pos = pf->getNextToVisit(objX, objY);
	real8 dist = distance(objX, objY, pos.x, pos.y);

	real8 p = 150 - dist * 0.1;
	if (p < 0) {
		p = 0;
	}
	return p;
}

real8 TotalPF::getBuildBaseP(sint4 objX, sint4 objY, uint4 unitId, const Game& game, PF* pf, bool useTrail) {
	if (objX < 0 || objY < 0 || objX > 1023 || objY > 1023) {
		return -1;
	}

	const Map<GameTile> &gameMap = game.get_map();
	int mX = objX / game.get_tile_points();
	int mY = objY / game.get_tile_points();
	const GameTile &tile = gameMap(mX, mY);
	if (tile.get_typeW() == TILE_UNKNOWN) {
		return 0;
	}
	
	real8 p = getTerrainCharge(objX, objY, Blackboard::TYPE_WORKER, true, game);// * 0.5;
	
	if (p <= TotalPF::CHARGE_IMPASSABLE_TILES) {
		return p;
	}

	p += pf->getBuilderBaseCharge(objX, objY, unitId, game.tick(), useTrail);
	p += pf->getBuilderEdgeCharge(objX, objY);
	
	return p;
}

real8 TotalPF::getTerrainCharge(sint4 objX, sint4 objY, int unitType, bool isBuilder, const Game& game) {
	const Map<GameTile> &gameMap = game.get_map();

	int mX = objX / game.get_tile_points();
	int mY = objY / game.get_tile_points();

	int baseRadius = 3; //Worker radius
	if (unitType == Blackboard::TYPE_MARINE) {
		baseRadius = 4;
	}
	if (unitType == Blackboard::TYPE_TANK) {
		baseRadius = 7;
	}
	
	real8 dist = 10000; //Default/max dist

	for (int x = mX - 3; x <= mX + 3; x++) {
		for (int y = mY - 3; y <= mY + 3; y++) {
			if (x >= 0 && y >= 0 && x < Blackboard::map_w && y < Blackboard::map_h) {

				const GameTile &tile = gameMap(x, y);

				if (tile.get_split() == Tile::NO_SPLIT) {
					
					//No split. Tile only has one terrain type
					if (tile.get_typeW() == TILE_CLIFF) {
						real8 tmp = getDistFull(objX, objY, x, y,  game.get_tile_points());
						if (tmp < dist) {
							dist = tmp;
						}
					}
				}
				else if (tile.get_split() == Tile::BT_SPLIT) {
					//Tile is split |\|
					if (tile.get_typeW() == TILE_CLIFF) {
						real8 tmp = getDistSplit(objX, objY, x, y,  game.get_tile_points());
						if (tmp < dist) {
							dist = tmp;
						}
					}
					else if (tile.get_typeE() == TILE_CLIFF) {
						real8 tmp = getDistSplit(objX, objY, x, y,  game.get_tile_points());
						if (tmp < dist) {
							dist = tmp;
						}
					}
					else {
						cout << "Unknown BT split" << endl;
					}
				}
				else if (tile.get_split() == Tile::TB_SPLIT) {
					//Tile is split |/|
		
					if (tile.get_typeW() == TILE_CLIFF) {
						real8 tmp = getDistSplit(objX, objY, x, y,  game.get_tile_points());
						if (tmp < dist) {
							dist = tmp;
						}
					}
					else if (tile.get_typeE() == TILE_CLIFF) {
						real8 tmp = getDistSplit(objX, objY, x, y,  game.get_tile_points());
						if (tmp < dist) {
							dist = tmp;
						}
					}
					else {
						cout << "Unknown TB split" << endl;
					}
				}
				else {
					cout << "Unknown tile split" << endl;
				}
			}
		}
	}

	real8 p = 0;

	if (!isBuilder) {
		if (dist <= baseRadius + 1) {
			p = TotalPF::CHARGE_IMPASSABLE_TILES;
		}
		else if (dist >= baseRadius + 11) {
			return 0;
		}
		else {
			p = -40 / ((dist/8)*(dist/8));// * TBConfig::params.Terrain_w;
		}
	}
	else {
		//62 (base width)
		if (dist <= 125) {
			p = TotalPF::CHARGE_IMPASSABLE_TILES;
		}
		else {
			p = 0;
		}
	}
	
	return p;
}

real8 TotalPF::getDistFull(sint4 objX, sint4 objY, int tX, int tY, int tilePts) {
	int tX1 = tX * tilePts;
	int tY1 = tY * tilePts;
	int tX2 = tX1 + tilePts;
	int tY2 = tY1 + tilePts;

	if (objX >= tX1 && objX <= tX2) {
		//inside X
		real8 d1 = distance(objX, objY, objX, tY1);
		real8 d2 = distance(objX, objY, objX, tY2);
		if (d1 < d2) {
			return d1;
		}
		else {
			return d2;
		}
	}

	if (objY >= tY1 && objY <= tY2) {
		//inside Y
		real8 d1 = distance(objX, objY, tX1, objY);
		real8 d2 = distance(objX, objY, tX2, objY);
		if (d1 < d2) {
			return d1;
		}
		else {
			return d2;
		}
	}

	real8 d1 = distance(objX, objY, tX1, tY1);
	real8 d2 = distance(objX, objY, tX1, tY2);
	real8 d3 = distance(objX, objY, tX2, tY1);
	real8 d4 = distance(objX, objY, tX2, tY2);

	real8 d = d1;
	if (d2 < d) {
		d = d2;
	}
	if (d3 < d) {
		d = d3;
	}
	if (d4 < d) {
		d = d4;
	}

	return d;
}

real8 TotalPF::getDistSplit(sint4 objX, sint4 objY, int tX, int tY, int tilePts) {
	int tX1 = tX * tilePts + tilePts / 2;
	int tY1 = tY * tilePts + tilePts / 2;

	return distance(objX, objY, tX1, tY1);
}

/*
	Returns the distance between two points (x1,y1) and (x2,y2).
*/
real8 TotalPF::distance(sint4 x1, sint4 y1, sint4 x2, sint4 y2) {
	//Correct calculation. Requires lot of CPU time.
	/*real8 ASquare = ((real8)x2 - (real8)x1) * ((real8)x2 - (real8)x1);
	real8 BSquare = ((real8)y2 - (real8)y1) * ((real8)y2 - (real8)y1);
	real8 D = sqrt(ASquare + BSquare);
	return D;*/

	//Optimized version. Requires less CPU time.
	real8 lX = (real8)abs(x2 - x1);
	real8 lY = (real8)abs(y2 - y1);
	real8 max = lX;
	real8 min = lY;
	if (lY > max) {
		max = lY;
		min = lX;
	}

	int t = (int)(min / max * (real8)100);
	real8 appD = max * preCalc[t];
	return appD;
}

real8 TotalPF::MHdistance(sint4 x1, sint4 y1, sint4 x2, sint4 y2) {
	return abs(x1 - x2) + abs(y1 - y2);
}

int TotalPF::getRealWidth() {
	return 1024; //TODO: Update to correct
}

int TotalPF::getRealHeight() {
	return 1024;
}
