#ifndef TotalPF_H
#define TotalPF_H

#include "PF.H"
#include "GameStateModule.H"
#include "GameObj.H"
#include "Game.H"
#include "Map.H"
#include "EventTimer.H"

using namespace std;

/*
	This class is responsible for calculating the total potential in a point for a
	specific unit.
*/
class TotalPF {

	typedef GameStateModule* GameStateModulePtr;

	public:

		/*
			Constructor.
		*/
		TotalPF(GameStateModulePtr mState);

		/*
			Destructor.
		*/
		~TotalPF();
		
		//The charge used for impassable terrain tiles
		static const int CHARGE_IMPASSABLE_TILES;
		
		/*
			Calculates the total potential in the point (objX,objY) for a military unit.
		*/
		real8 getUnitP(sint4 objX, sint4 objY, bool offensive, uint4 unitId, int unitType, const Game& game, PF* pf);

		/*
			Calculates the total potential in the point (objX,objY) for a military unit.
		*/
		real8 getWorkerP(sint4 objX, sint4 objY, sint4 currentState, uint4 unitId, int unitType, const Game& game, PF* pf, bool useTrail);

		/*
			Calculates the potential the terrain generates in point (objX,objY).
		*/
		real8 getTerrainCharge(sint4 objX, sint4 objY, int unitType, bool isBuilder, const Game& game);

		/*
			Calculates distance to a fully impassable terrain tile.
			Help method for terrain calculation.
		*/
		real8 getDistFull(sint4 objX, sint4 objY, int tX, int tY, int tilePts);

		/*
			Calculates distance to a partially (splitted) impassable terrain tile.
			Help method for terrain calculation.
		*/
		real8 getDistSplit(sint4 objX, sint4 objY, int tX, int tY, int tilePts);

		/*
			Calculates the potential explorer units has in point (objX,objY)
		*/
		real8 getUnexplCharge(sint4 objX, sint4 objY, uint4 unitId, PF* pf);

		/*
			Calculates the potential for building structure in point (objX,objY). Note that this potential field
			is used to find a buildspot, not for navigation.
		*/
		real8 getBuildSpotCharge(sint4 objX, sint4 objY, uint4 unitId, PF* pf);
		
		/*
			Calculates the navigation potential for getting to a previously found buildspot.
		*/
		real8 getBuildBaseP(sint4 objX, sint4 objY, uint4 unitId, const Game& game, PF* pf, bool useTrail);

		/*
			Returns the distance between two points (x1,y1) and (x2,y2).
		*/
		static real8 distance(sint4 x1, sint4 y1, sint4 x2, sint4 y2);

		/*
			Returns the Manhatten distance (abs(x2-x1)+abs(y2-y1)) between two points (x1,y1) and (x2,y2).
		*/
		static real8 MHdistance(sint4 x1, sint4 y1, sint4 x2, sint4 y2);

		/*
			Returns the map width in points.
		*/
		int getRealWidth();

		/*
			Returns the map height in points.
		*/
		int getRealHeight();

	private:

		static const real8 preCalc[101];

};

#endif
