#ifndef UnitAgent_H
#define UnitAgent_H

#include "TotalPF.H"
#include "PFUpdater.H"
#include "GameObj.H"
#include "Game.H"
#include "GameStateModule.H"
#include "FireCoordinator.H"

/*
	This class represents an autonomous agent for units. It compute actions
	for a tank or marine based on the current game state, potential field and firecoordinator.
*/
class UnitAgent {
	
	typedef TotalPF* TotalPFPtr;
	typedef GameObj* GameObjPtr;
	typedef PFUpdater* PFUpdaterPtr;
	typedef GameStateModule* GameStateModulePtr;
	typedef FireCoordinator* FireCoordinatorPtr;

	public:

		/*
			Main "execute action" method for each unit. Call this method each frame.
		*/
		static void compute_actions(GameObjPtr unit, GameStateModulePtr state, PFUpdaterPtr pfupdater, FireCoordinatorPtr coordinator, Random* rand);

		/*
			Decides which mode (offensive = true, defensive = false) to use for a unit.
		*/
		static bool getMode(GameObjPtr unit, GameStateModulePtr state);
		
		/*
			Check if the straight path from (fromX,fromY) to (toX,toY) is blocked or not. Currently not working correctly.
		*/
		static bool canDoMove(uint4 objId, int fromX, int fromY, int toX, int toY, int unitType, PFUpdaterPtr pfupdater, const Game& game);

		/*
			Help method for canDoMove. Used to prevent units getting stack at map edges.
		*/
		static bool boundaryMove(int fromX, int fromY, int toX, int toY, TotalPFPtr totalpf);

		/*
			Help method for canDoMove. Calculates the next position against a straight path.
		*/
		static int calcNext(int cnt, int diffVal, int divider);

		/*
			Checks for enemy targets within range and fires if any targets found.
			Returns true if fire at a target.

			This method uses a global target coordination to decide what target to attack (if more than one is within firerange).
		*/
		static bool canFire(GameObjPtr unit, int unitX, int unitY, uint4 objId, GameStateModulePtr state, bool offensive, FireCoordinatorPtr coordinator);

		/*
			Checks if there are better positions nearby to move to.
			Returns true if unit moves, false if not.
		*/
		static bool canMove(GameObjPtr unit, int unitX, int unitY, uint4 objId, bool offensive, PFUpdaterPtr pfupdater, const Game& game);
		
		//Max firerange for tanks
		static const int TANK_FIRERANGE;
		//Max firerange for marines
		static const int MARINE_FIRERANGE;
		//Max HP for tanks
		static const int TANK_HP;
		//Armor for tanks
		static const int TANK_ARMOR;
		//Max HP for marines
		static const int MARINE_HP;
		//Armor for marins
		static const int MARINE_ARMOR;
		//Max HP for marines
		static const int WORKER_HP;
		//Armor for marins
		static const int WORKER_ARMOR;
		//Max HP for control centers
		static const int CONTROLCENTER_HP;
		//Armor for control centers
		static const int CONTROLCENTER_ARMOR;
		//Max HP for barracks
		static const int BARRACKS_HP;
		//Armor for barracks
		static const int BARRACKS_ARMOR;
		//Max HP for factory
		static const int FACTORY_HP;
		//Armor for factory
		static const int FACTORY_ARMOR;

		//Min/max damage for tank attacks
		static const sint4 TANK_MIN_DMG;
		static const sint4 TANK_MAX_DMG;
		//Min/max damage for marine attacks
		static const sint4 MARINE_MIN_DMG;
		static const sint4 MARINE_MAX_DMG;
};

#endif
